import osmnx as ox
import networkx as nx
import folium
import itertools
import pickle
import time, datetime
import os.path

import random

import os
import glob


from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.ops import linemerge

import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)

start_time = time.time()

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length

def get_bounds(Q):
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    tb = edges.total_bounds
    print( tb )
    sw = [tb[1],tb[0]]
    ne = [tb[3],tb[2]]
    return [sw, ne]

def save_png(Q, max_Q=None, bounds=None, index = 0):
    m = folium.Map()
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    length = graph_length(Q)
    iteration_label = 'iteration '+ str(index) + ' ' + str(round(length)) + ' miles'
    ax = edges.plot(color='#0000ff', alpha=0.2, zorder=1, label = iteration_label)

    if max_Q != None:
        QD = max_Q.to_directed()
        QD.graph['crs'] = ox.settings.default_crs
        nodes, edges = ox.graph_to_gdfs(QD)
        max_length = graph_length(max_Q)
        max_label = 'current max: ' + str(round(max_length)) + ' miles'            
        edges.plot(ax=ax, color='#ff0000', alpha=1.0 , zorder=0, linestyle='dotted', label = max_label)
    if bounds != None:
        sw = bounds[0]
        ne = bounds[1]
        ax.set_xlim(sw[1], ne[1])
        ax.set_ylim(sw[0], ne[0])
    ax.legend(loc="lower left")
    ax.plot()
    plt.axis('off')
    plt.savefig('fig/img{}.png'.format(index), dpi=120)
    plt.close()
 
  
def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('')
        

H = pickle.load(open('h_graph.pkl', 'rb'))
report(H,'graph loaded from disk')

# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(H), key=len)
J = nx.subgraph(H, nodes).copy()
#report(J,'in largest subgraph')
#draw(J, 'largest subgraph')

# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

delete_ends(J)

T = J.to_directed()
T = ox.simplification.simplify_graph(T, strict=True, remove_rings=True, track_merged=True)
J = T.to_undirected()
        
# get a graph of just the odd nodes, improve later to handle 4-way intersections

def odd_nodes_graph(Q):
    R = Q.copy()
    unwanted_edges = []
    for u, v, d in R.edges(data=True):
        if R.degree(u) % 2 != 1 or R.degree(v) % 2 != 1:
            unwanted_edges.append( (u,v) )
    R.remove_edges_from( unwanted_edges )
    return R
    

K = odd_nodes_graph(J)           
# add in edges that bypass 4-nodes in graph
# fix: 4-way connected to a 4-way
def complete_3node_graph(Q,R):
    n4list = []
    for node, neighbors in Q.adjacency():
        if len(neighbors) == 4:
            n4list.append( (node,neighbors) )
            nlist = []
            llen = []
            for neighbor, edict in neighbors.items():
                w = min([d['length'] for d in edict.values()])
                nlist.append(neighbor)
                llen.append(w)
              #  for dic in edict.items():
                    #w = dic['length']
                    #nlist.append(neighbor)
                    #llen.append(w)
            # could use itertools?
            for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:
                R.add_edge(nlist[i],nlist[j], length= llen[i]+llen[j]) 
    return n4list

fix_4_list = complete_3node_graph(J,K)

report(K,'in odd-nodes 4node bypass')

def restore_4node_graph(Q,fix_4_list):
    for node, neighbors in fix_4_list:
       i = 0
       nlist = []
       vlist = []
       for neighbor, edict in neighbors.items():
           # for dic in edict:
           nlist.append(neighbor)
           vlist.append(edict[0])
       for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:               
            if Q.has_edge(nlist[i],nlist[j]):
                Q.remove_edge(nlist[i],nlist[j])
                if not Q.has_edge(node,nlist[i]):
                    Q.add_edge(node,nlist[i],**vlist[i]) #fix
                if not Q.has_edge(node,nlist[j]):                                  
                    Q.add_edge(node,nlist[j],**vlist[j])

all_matchings = []
max_length = 0

# delete png files from any previous run
files = glob.glob('fig/' + 'img*.png')
for f in files:
    os.remove(f)

bounds = get_bounds(K)
max_T2 = None
max_length = 0
max_fuzzed_Graph = None

matching = nx.min_weight_matching(nx.Graph(K),'length')
print('length of min_weight_matching = ', len(matching))
print('is perfect matching? ', nx.is_perfect_matching(K,matching))
T = K.copy()
T.remove_edges_from(matching)
restore_4node_graph(T,fix_4_list)
nodes = max(nx.connected_components(T), key=len)
T2 = nx.subgraph(T, nodes).copy()
if nx.is_eulerian(T2):
    all_matchings.append(matching)
    report(T2, 'min_weight matching')
    print('length of successful matching: ', len(matching))
    print('min_weight matching is eulerian')
    length = graph_length(T2)
    all_matchings.append(matching)
    index = len(all_matchings)-1
    save_png(T2,max_T2,bounds,index)                                             
    if length > max_length:
        max_length = length
        max_T2 = T2.copy()

matching = nx.min_weight_matching(nx.Graph(K))
print('length of no-weight min_weight_matching = ', len(matching) )
print('is no-weight matching perfect? ', nx.is_perfect_matching(K,all_matchings[1]))
T = K.copy()
T.remove_edges_from(matching)
restore_4node_graph(T,fix_4_list)
nodes = max(nx.connected_components(T), key=len)
T2 = nx.subgraph(T, nodes).copy()
if nx.is_eulerian(T2):
    all_matchings.append(matching)
    print('length of successful matching: ', len(matching))
    print('no-weight matching is eulerian')
    length = graph_length(T2)
    index = len(all_matchings)-1
    save_png(T2,max_T2,bounds,index)
    if length > max_length:
        max_length = length
        max_T2 = T2.copy()
else:    
    print('no-weight matching is NOT eulerian')
report(T2, 'matching[1]')


def fuzz_min_weight_matching(Q):
    T = nx.Graph()
    for u, v, d in Q.edges(data=True):
        len = d['length']
        len = len * (1 + random.random())
        T.add_edge(u,v,length=len)
    return T, nx.min_weight_matching(T,weight='length')

j=0
while j<1000:
    if max_fuzzed_Graph == None:
        fuzzed_Graph, matching = fuzz_min_weight_matching(K)
    else:
        fuzzed_Graph, matching = fuzz_min_weight_matching(max_fuzzed_Graph)
    j+=1
    T = K.copy()
    T.remove_edges_from(matching)
    restore_4node_graph(T,fix_4_list)
    nodes = max(nx.connected_components(T), key=len)
    T2 = nx.subgraph(T, nodes).copy()
    if nx.is_eulerian(T2):
        all_matchings.append(matching)
        print('found eulerian on fuzz attempt ', j)
        index = len(all_matchings)-1
        print('length of successful matching: ', len(matching))
        print('matching ', index,  ' is eulerian')
        report(T2,'matching ' + str(index))
        length = graph_length(T2)
        save_png(T2,max_T2,bounds,index)                                               
        if length > max_length:
            max_length = length
            max_T2 = T2.copy()
            max_fuzzed_Graph = fuzzed_Graph.copy()
        print('current max length: ',round(max_length))
        pickle.dump(max_T2,open('euler_graph.pkl', 'wb'))

                                                
    
elapsed_time(start_time)
print('finished finding some possible matchings')
print('max length = ', max_length, ' miles')


pickle.dump(all_matchings,open('all_matchings.pkl', 'wb'))
pickle.dump(fix_4_list,open('fix_4_list.pkl', 'wb'))
pickle.dump(K,open('k_graph.pkl', 'wb'))
pickle.dump(max_T2,open('maxt2_graph.pkl', 'wb'))


elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






